from __future__ import annotations
from .verifyta import Verifyta
# import os



def set_verifyta_path(verifyta_path: str):
    v = Verifyta()
    v.set_verifyta_path(verifyta_path)

# def simple_verify(model_path: str | List[str], 
#                   trace_path: str | List[str], 
#                   parallel: str=None):
#     v = Verifyta()
#     return v.simple_verify(model_path, trace_path, parallel)


    


from typing import List, Tuple, Dict
import xml.etree.cElementTree as ET

class UFactory:
    """
    usage: UFactory.function(xxx)
    用户一般不直接调用这个class
    用来生产UPPAAL里的xml element组件，主要包含
    declaration, (简单字符串)
    query, (简单字符串)
    queries, (多个queries)
    location,
    transition,
    template,
    """

    @staticmethod
    def declaration(declaration: str):
        """
        构造UPPAAL的declaration
        declaration: str, declaration里的全部内容
        """
        dec_elem = ET.Element('declaration')
        dec_elem.text = declaration
        return dec_elem

    @staticmethod
    def __query(query: str):
        """
        query: str, 验证语句
        构造一个query
        query示例：
            <query>
                <formula>E&lt;&gt; Output.pass</formula>
                <comment />
            </query>
        """
        query_elem = ET.Element('query')

        # 构建并添加formula
        formula_elem = ET.Element('formula')
        formula_elem.text = query
        query_elem.append(formula_elem)

        # 构建并添加comment
        query_elem.append(ET.Element('comment'))
        return query_elem

    @staticmethod
    def queries(queries: List[str]):
        """
        queries: List[str], 多条验证语句
        完整queries示例
        <queries>
            <query>
                <formula>E&lt;&gt; Output.pass</formula>
                <comment />
            </query>
            <query>
                <formula>E&lt;&gt; Output2.pass</formula>
                <comment />
            </query>
        </queries>
        """
        queries_elem = ET.Element('queries')
        # 构建并加入多个queries element
        for query in queries:
            queries_elem.append(UFactory.__query(query))
        return queries_elem

    @staticmethod
    def location(location_id: int, pos_x: int, pos_y: int,
                 inv: str = None, name: str = None, is_committed: bool = False):
        """
        构造一个location
        id: int, location的id，注意不要和以前的location的id重复
        pos_x, pos_y: int, location的位置
        inv: str, location的invariant，比如'gclk<=10'或者'gclk<100'
        name: str, location的名字，默认是None
        is_committed: bool, 是否为committed，默认不是committed

        完整location示例：
            <location id="id37" x="-169" y="-59">
                <name x="90" y="166">pass</name>
                <label kind="invariant" x="-212" y="-42">gclk&lt;=122</label>
                <committed/>
            </location>
        """
        location = ET.Element('location', {'id': f'id{location_id}',
                                           'x': str(pos_x),
                                           'y': str(pos_y)})
        # 添加名字
        if name is not None:
            location_name = ET.Element('name', {'x': str(pos_x),
                                                'y': str(pos_y - 20)})
            location_name.text = name
            location.append(location_name)
        # 添加inv
        if inv is not None:
            label = ET.Element('label', {'kind': 'invariant',
                                         'x': str(pos_x),
                                         'y': str(pos_y - 40)})
            label.text = inv
            location.append(label)
        # 添加committed
        if is_committed:
            location.append(ET.Element('committed'))
        return location

    @staticmethod
    def transition(sourceID: int, targetID: int, pos_x: int, pos_y: int, 
                   guard: str = None, sync: str = None, clock_reset: str = None, 
                   nail: bool = False):
        """
        sourceID: int, 起点id
        targetID: int, 终点id
        post_x, post_y: int, 符号的位置
        guard: str, edge的guard，如't>=10', 't>100'
        sync: str, synchronisation信号，需要注明!或者？
        clock_reset: str, 如't=0'
        nail: bool, False: 表示是否让边弯曲，适用于同起点终点存在多条路径的情况。
        完整transition示例：
        <transition>
            <source ref="id7"/>
            <target ref="id6"/>
            <label kind="guard" x="139" y="90">t&gt;=tERPMin</label>
            <label kind="synchronisation" x="123" y="90">sync?</label>
            <label kind="assignment" x="156" y="107">t=0</label>
        </transition>
        """
        transition = ET.Element('transition')
        # 构建并添加source
        source = ET.Element('source', {'ref': f'id{sourceID}'})
        transition.append(source)
        # 构建并添加target
        target = ET.Element('target', {'ref': f'id{targetID}'})
        transition.append(target)
        # 构建并添加guard
        if guard is not None:
            label_guard = ET.Element('label', {'kind': 'guard',
                                               'x': str(pos_x),
                                               'y': str(pos_y)})
            label_guard.text = guard
            transition.append(label_guard)
        # 构建并添加synchronisation
        if sync is not None:
            label_sync = ET.Element('label', {'kind': 'synchronisation',
                                              'x': str(pos_x),
                                              'y': str(pos_y - 30)})
            label_sync.text = sync
            transition.append(label_sync)
        # 构建并添加assignment
        if clock_reset is not None:
            label_assignment = ET.Element('label', {'kind': 'assignment',
                                                    'x': str(pos_x),
                                                    'y': str(pos_y - 60)})
            label_assignment.text = clock_reset
            transition.append(label_assignment)
        # 构建并添加nail (弯曲结点)
        if nail:
            nail_element = ET.Element('nail', {'x': str(pos_x),
                                               'y': str(pos_y)})
            transition.append(nail_element)
        return transition

    @staticmethod
    def template(name: str, locations: List[ET.Element], init_id: int, transitions: List[ET.Element],
                 parameter: str = None, declaration: str = None):
        """
        name: str, template的名字
        locations: List[ET.Element], 由UFactory.construct_location()构造出的一系列locations
        init_id: int, initial location的id，比如47
        transitions: List[ET.Element], 由UFactory.construct_transition()构造出的一系列transitions
        parameter: str,
            比如：<parameter>broadcast chan &amp;actNode, actPath,int tERPMin, int tERPMax</parameter>
        declaration: str,
            比如: <declaration>clock t;</declaration>
        构建一个新的template
        一个template的基本框架为：
            <template>
                <name>NodeN</name>
                可以没有<parameter>broadcast chan &amp;actNode, int tERPMin, int tERPMax</parameter>
                可以没有<declaration>clock t;</declaration>

                <location> </location>
                <location> </location>
                <location> </location>
                <init ref="id47"/>
                <transition> </transition>
                <transition> </transition>
                <transition> </transition>
            </template>
        """
        template = ET.Element('template')
        # 创建并加入name
        name_elem = ET.Element('name')
        name_elem.text = name
        template.append(name_elem)

        # 创建并加入parameter
        if parameter is not None:
            parameter_elem = ET.Element('parameter')
            parameter_elem.text = parameter
            template.append(parameter_elem)

        # 创建并加入declaration
        if declaration is not None:
            declaration_elem = ET.Element('declaration')
            declaration_elem.text = declaration
            template.append(declaration_elem)

        # 加入locations
        for location in locations:
            template.append(location)

        # 创建并加入init
        init_elem = ET.Element('init', {'ref': f'id{init_id}'})
        template.append(init_elem)

        # 加入 transitions
        for transition in transitions:
            template.append(transition)
        return template

    @staticmethod
    def input(input_name: str, signals: List[Tuple[str, str, str]], init_id: int):
        """
        构建输入的线性模型，类似monitor
        input_name: string
        signals: List[Tuple[str, int, int]],
                 每个tuple分别对应[signal, guard, inv]
                 signal, guard, inv, name 都可以是None
                 signal: 信号名称
                 guard: str, 比如gclk>=10
                 inv: str, 比如gclk<=5
                 注意：在连接的时候是按照信号在list出现的顺序连接的
        init_id: int, 用来设置Monitor中最小的<init ref='xxx'/>
        """
         # 创建locations
        locations = []
        for i in range(len(signals)):
            # [signal, guard, inv, name]
            location = UFactory.location(init_id + i, 300 * i, 200, signals[i][2])
            locations.append(location)
        # 需要多一个尾巴location
        location = UFactory.location(init_id + len(signals), 300 * len(signals), 200, name='pass')
        locations.append(location)
        # 创建transitions
        transitions = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            transition = UFactory.transition(init_id + i, init_id + i + 1, i * 300 + 100, 200,
                                                  signals[i][1], signals[i][0]+'!')
            transitions.append(transition)
        # 获得clock name并创建declaration
        clk_name = signals[0][1].split('>')[0]
        declaration = f'clock {clk_name};'
        monitor_elem = UFactory.template(input_name, locations, init_id, transitions,declaration=declaration)
        return monitor_elem

    @staticmethod
    def monitor(monitor_name: str, signals: List[Tuple[str, str, str]], observe_action: List[str], 
                init_id: int, strict: bool = False, allpattern: bool=False):
        """
        构建新的线性monitor
        monitor_name: string
        signals: List[Tuple[str, int, int]],
                 每个tuple分别对应[signal, guard, inv]
                 signal, guard, inv, name 都可以是None
                 signal: 信号名称
                 guard: str, 比如gclk>=10
                 inv: str, 比如gclk<=5
                 注意：在连接的时候是按照信号在list出现的顺序连接的
        init_id: int, 用来设置Monitor中最小的<init ref='xxx'/>
        strict: bool, 是否是严格观测。如果strict是True，则要求在给定观测之外禁止有其他观测。
        附：一个Monitor的基本框架为：
            <template>
                <name>Monitor</name>
                <location> </location>
                <init ref="id47"/>
                <transition> </transition>
            </template>
        """
        # 创建locations
        locations = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            inv = None if allpattern else signals[i][2]
            location = UFactory.location(init_id + i, 300 * i, 200, inv=inv)
            locations.append(location)
        # 需要多一个尾巴location
        location = UFactory.location(init_id + len(signals), 300 * len(signals), 200, name='pass')
        locations.append(location)

        # 创建transitions
        transitions = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            guard = None if allpattern else signals[i][1]
            transition = UFactory.transition(init_id + i, init_id + i + 1, i * 300 + 100, 200,
                                                  guard=guard, sync=signals[i][0]+'?')
            transitions.append(transition)

        # 如果是strict，需要给每个location创建fail location，并且构建对应transitions边指向对应的fail location
        if strict:
            # 构建指向fail的transitions
            for i in range(len(signals)):
                # 构建一个fail location
                fail_location_id = init_id + len(signals) + i + 1
                # check allpattern
                inv = None if allpattern else signals[i][2].replace('<=', '<')
                location = UFactory.location(fail_location_id, 300 * i, -200, inv=inv,
                                                  name='fail' + str(i))
                locations.append(location)
                actions_length = 200 // len(observe_action)
                # 需要对每一个observe action指向fail transitions
                for j in range(len(observe_action)):
                    # 指向fail，注意guard是大于
                    guard = None if allpattern else signals[i - 1][1].replace('>=', '>')
                    if allpattern and (observe_action[j] == signals[i][0]):
                        continue
                    if i == 0:
                        transition = UFactory.transition(init_id + i, fail_location_id, i * 300 + (j-len(observe_action)//2) * actions_length, -100,
                                                         guard=None, sync=observe_action[j]+'?',nail=True)
                    else:
                        transition = UFactory.transition(init_id + i, fail_location_id, i * 300 + (j-len(observe_action)//2) * actions_length, -100,
                                                         guard=guard, sync=observe_action[j]+'?',nail=True)

                    transitions.append(transition)
        # 获得clock name并创建declaration
        clk_name = signals[0][1].split('>')[0]
        declaration = None if allpattern else f'clock {clk_name};'
        monitor_elem = UFactory.template(monitor_name, locations, init_id, transitions, declaration=declaration)
        return monitor_elem

    @staticmethod
    def signal_converter(signal_dict: Dict[str, str], init_id: int, converter_name: str = 'SignalConverter'):
        """
        将某些可观测信号转化成另一些信号
        signal_dict: Dict[str, str]，信号对应的字典
        init_id: int, location最小的id，防止location的冲突

        比如：signal_dict = {'actPathHisA': 'sigOut',
                            'actPathHisH': 'sigOut',
                            'actPathHisV': 'sigOut',
                            'sigIn': 'actNodeSA'}
        会构造出如xx图片的自动机
        """
        pass



# support typing str | List[str]
# https://github.com/microsoft/pylance-release/issues/513
from __future__ import annotations
import warnings
from typing import List
import os
from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool
import functools
from multiprocessing import cpu_count
import platform

is_windows = platform.system() == 'Windows'

def check_is_verifyta_path_empty(func):
    """
    装饰器，用来在verifyta运行前检测路径是否被设置。
    """
    @functools.wraps(func)
    def checker_wrapper(*args, **kwargs):
        # 注意verifyta是singleton，因此可以直接用Verifyta()调用到唯一的实例
        if Verifyta()._verifyta_path:
            return func(*args, **kwargs)
        else:
            error_info = 'Verifyta path is not set.'
            error_info += ' Please use "Verifyta().set_verifyta_path(verifyta_path: str)" to set the path of verifyta.'
            raise ValueError(error_info)
    return checker_wrapper


class Verifyta:
    """
    uppaal commandline tools
    """
    # make singleton
    _instance = None
    _is_first_init = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        # make singleton
        if not self._is_first_init:
            return
        self._is_first_init = False

        self._verifyta_path: str = ''

    @property
    def is_verifyta_empty(self):
        """
        check whether verifyta is set
        """
        if self._verifyta_path:
            return True
        else:
            return False

    @property
    def is_valid_verifyta_path(self):
        res = os.popen(f'{self._verifyta_path} -h').read()
        if 'Usage: verifyta [OPTION]... MODEL QUERY' in res:
            return True
        else:
            return False

    def set_verifyta_path(self, verifyta_path: str):
        """
        设置verifyta_path
        verifyta_path: str, 用户指定的verifyta_path

        设置时会验证verifyta path的合法性，方法如下：
        通过命令行调用f'{verifyta_path} -h'，查看返回内容是否包含 'Usage: verifyta [OPTION]... MODEL QUERY'
        """
        res = os.popen(f'{verifyta_path} -h').read()
        if '-h [ --help ]' in res:
            self._verifyta_path = verifyta_path
        else:
            example_info = "======== Example Paths ========" \
                           "\nWindows: absolute_path_to_uppaal\\bin-Windows\\verifyta.exe" \
                           "\nLinux  : absolute_path_to_uppaal/bin-Linux/verifyta" \
                           "\nmacOS  : absolute_path_to_uppaal/bin-Darwin/verifyta"
            raise ValueError(f"Invalid verifyta_path: {verifyta_path}.\n{example_info}")

    def get_verifyta_path(self):
        return self._verifyta_path

    @check_is_verifyta_path_empty
    def simple_verify(self, 
                      model_path: str | List[str], 
                      trace_path: str | List[str] = None, 
                      parallel: str=None, options = None):
        """
        Simple verification, return to the shortest diagnostic path. 
        Verify the model in model_path and save the verification results to trace_path
        
        model_path: str or str list, Model paths to be verified
        trace_path: str or str list, Trace paths to be saved
        parallel: str, select parallel method for accelerate verification, 
        None(default):run in sequence, 'process':use multiprocessing, 'threads': use multithreads.
        """
        # check trace_path is None
        if trace_path is None:
            if type(model_path) == str:
                trace_path = os.path.splitext(model_path)[0] + '.xtr'
            else:
                trace_path = [os.path.splitext(x)[0]+'.xtr' for x in model_path]

        # check model_path and trace_path type is same
        if type(model_path) != type(trace_path):
            error_info = f'type of model_path and trace_path are inconsistent\n'
            error_info += f'model_path: {type(model_path)}, trace_path:{type(trace_path)}.'
            raise TypeError(error_info)

        # check model_path is only one model, set parallel loop
        if type(model_path) != list:
            model_path, trace_path = [model_path], [trace_path]
            parallel = None
        
        # check model_path and trace_path len is same
        if len(model_path) != len(trace_path):
            error_info = f'length of model_path and trace_path are inconsistent'
            raise ValueError(error_info)
        
        # set uppaal environment variables
        cmd_env = 'set UPPAAL_COMPILE_ONLY=' if is_windows else "UPPAAL_COMPILE_ONLY="
        
        cmds = []
        for i in range(len(model_path)):
            model_i = model_path[i]
            trace_i = trace_path[i]
            # check model_path exist
            if not os.path.exists(model_i):
                error_info = f'model_path {model_i} not found.'
                raise FileNotFoundError(error_info)
            # check trace_path format
            if trace_i.endswith('.xml'):
                trace_i = trace_i.replace('.xml', '')
                if options is not None:
                    options_str = " ".join(options)
                    cmd = cmd_env+'&&'+f'{self._verifyta_path} -X {trace_i} {options_str} {model_i}'
                else:
                    cmd = cmd_env+'&&'+f'{self._verifyta_path} -t 1 -X {trace_i} {model_i}'
                cmds.append(cmd)
            elif trace_i.endswith('.xtr'):
                trace_i = trace_i.replace('.xtr', '')
                if options is not None:
                    options_str = " ".join(options)
                    cmd = cmd_env+'&&'+f'{self._verifyta_path} -f {trace_i} {options_str} {model_i}'
                else:
                    cmd = cmd_env+'&&'+f'{self._verifyta_path} -t 1 -f {trace_i} {model_i}'
                cmds.append(cmd)
            else:
                error_info = 'trace_path should end with ".xml" or ".xtr".'
                error_info += f' Currently trace_path = {trace_i}'
                raise ValueError(error_info)

        # select parallel method
        if parallel == None:
            return self.cmds_loop(cmds=cmds)
        if parallel == 'process':
            res = self.cmds_process(cmds=cmds)
        elif parallel == 'threads':
            res = self.cmds_threads(cmds=cmds)
        else:
            error_info = 'parallel should be "process" or "threads".'
            error_info += f' Currently parallel = {parallel}'
            raise ValueError(error_info)
        return res 

    # @check_is_verifyta_path_empty
    # def simple_verify_process(self, model_paths: List[str], trace_paths: List[str]):
    #     cmds = []
    #     for i in range(len(model_paths)):
    #         model_path = model_paths[i]
    #         trace_path = trace_paths[i]

    #         # check model_path exist
    #         if not os.path.exists(model_path):
    #             error_info = f'model_path {model_path} not found.'
    #             raise FileNotFoundError(error_info)

    #         # check trace_path format
    #         if trace_path.endswith('.xml'):
    #             trace_path = trace_path.replace('.xml', '')
    #             cmd = f'{self._verifyta_path} -t 1 -X {trace_path} {model_path}'
    #             cmds.append(cmd)
    #         elif trace_path.endswith('.xtr'):
    #             trace_path = trace_path.replace('.xtr', '')
    #             cmd = f'{self._verifyta_path} -t 1 -f {trace_path} {model_path}'
    #             cmds.append(cmd)
    #         else:
    #             error_info = 'trace_path should end with ".xml" or ".xtr".'
    #             error_info += f' Currently trace_path = {trace_path}'
    #             raise ValueError(error_info)

    #     return self.cmds_process(cmds, len(cmds))

    # @check_is_verifyta_path_empty
    # def simple_verify_threads(self, model_paths: List[str], trace_paths: List[str]):
    #     cmds = []
    #     for i in range(len(model_paths)):
    #         model_path = model_paths[i]
    #         trace_path = trace_paths[i]

    #         # check model_path exist
    #         if not os.path.exists(model_path):
    #             error_info = f'model_path {model_path} not found.'
    #             raise FileNotFoundError(error_info)

    #         # check trace_path format
    #         if trace_path.endswith('.xml'):
    #             trace_path = trace_path.replace('.xml', '')
    #             cmd = f'{self._verifyta_path} -t 1 -X {trace_path} {model_path}'
    #             cmds.append(cmd)
    #         elif trace_path.endswith('.xtr'):
    #             trace_path = trace_path.replace('.xtr', '')
    #             cmd = f'{self._verifyta_path} -t 1 -f {trace_path} {model_path}'
    #             cmds.append(cmd)
    #         else:
    #             error_info = 'trace_path should end with ".xml" or ".xtr".'
    #             error_info += f' Currently trace_path = {trace_path}'
    #             raise ValueError(error_info)

    #     return self.cmds_threads(cmds, len(cmds))


    @check_is_verifyta_path_empty
    def cmd(self, cmd: str):
        """
        run common command with cmd, you can easily ignore the verifyta path.
        return the running cmd and the command result
        """
        if self._verifyta_path not in cmd:
            cmd = f'{self._verifyta_path} {cmd}'
        return cmd, os.popen(cmd).read()

    @check_is_verifyta_path_empty
    def cmds_loop(self, cmds: List[str]):
        """
        run in sequence
        """
        return [self.cmd(tmp_cmd) for tmp_cmd in cmds]

    @check_is_verifyta_path_empty
    def cmds_process(self, cmds: List[str], num_process: int = None):
        """
        note that sometimes, multiprocess is not faster than single-process or multi-threads
        run a list of commands and return results
        if num_process is not given, it will run with num cpu cores
        if num_process is 1, it's better run with self.cmd
        return running cmds and associated result
        """
        if num_process == None:
            num_process = cpu_count()
        
        if num_process == 1:
            w = 'You are running with only 1 process, we recommend using Verifyta().cmd() method.'
            warnings.warn(w)
        p = Pool(min(num_process, cpu_count()))
        return p.map(self.cmd, cmds)

    @check_is_verifyta_path_empty
    def cmds_threads(self, cmds: List[str], num_threads: int = None):
        """
        run a list of commands and return results
        if num_threads is not given, it will run with num cpu cores * 2
        if num_threads is 1, it's better run with self.cmd
        return running cmds and associated result
        """
        if num_threads == None:
            num_threads = cpu_count()*2
        
        if num_threads == 1:
            w = 'You are running with only 1 thread, we recommend using Verifyta().cmd() method.'
            warnings.warn(w)
        p = ThreadPool(min(num_threads, cpu_count()*2))
        return p.map(self.cmd, cmds)

    @check_is_verifyta_path_empty
    def compile_to_if(self, model_path: str):
        """Compile model_path(model.xml) to generate a intermediate format file (model.if). 

        Args:
            model_path (str): str or str list, Model path to be verified

        Raises:
            FileNotFoundError: model_path not found.
            ValueError: model_path is not a xml format file.
        """
        if not os.path.exists(model_path):
            error_info = f'model_path {model_path} not found.'
            raise FileNotFoundError(error_info)
        file_path, file_ext = os.path.splitext(model_path)
        if_path = file_path + '.if'
        
        if file_ext != '.xml':
            error_info = f'model_path {model_path} should be xml format file.'
            raise ValueError(error_info)

        # set uppaal environment variables
        cmd_env = 'set UPPAAL_COMPILE_ONLY=1'
        cmd = cmd_env+"&&"+f'{self._verifyta_path} {model_path} > {if_path}'
        self.cmd(cmd=cmd)
        
        if not os.path.exists(if_path):
            error_info = f'if file {if_path} has not generated.'
            raise FileNotFoundError(error_info)
        return if_path

        


# coding=utf-8
from ast import Raise
import sys
from typing import List, Tuple, Dict
import xml.etree.cElementTree as ET

from pyuppaal.namedtuple import TimedActions
from .verifyta import Verifyta
from .iTools import UFactory
from .tracer import SimTrace, Tracer
import os

# verifyta_ins = Verifyta()

class UModel:
    """
    载入UPPAAL模型，进行分析、编辑、验证、保存等操作
    """

    def __init__(self, model_path: str):
        # 模型路径，如 '../AVNRT_Initial_straight.xml'
        self.model_path: str = model_path
        self.element_tree: ET.ElementTree = ET.ElementTree(file=self.model_path)
        self.root_elem: ET.Element = self.element_tree.getroot()
        if not Verifyta()._verifyta_path:
            raise ValueError('Path of verifyta is not set. Please do Verifyta().set_verifyta_path().')

    def get_model_path(self):
        return self.model_path

    def save(self):
        new_model_path = self.model_path
        with open(new_model_path, 'w') as f:
            self.element_tree.write(new_model_path, encoding="utf-8", xml_declaration=True)
        self.model_path = new_model_path
        return self.model_path

    def save_as(self, new_model_path: str):
        with open(new_model_path, 'w') as f:
            self.element_tree.write(new_model_path, encoding="utf-8", xml_declaration=True)
        self.model_path = new_model_path
        return new_model_path

    def copy_as(self, new_model_path: str):
        with open(new_model_path, 'w') as f:
            self.element_tree.write(new_model_path, encoding="utf-8", xml_declaration=True)
        # self.model_path = new_model_path
        return new_model_path
    
    def get_communication_graph(self, save_path=None):
        """
        get the communication graph of the uppaal model and save it to a `.md` file
        save_path: string, the path of aiming file
        return mermaid string
        """
        from iTools import build_cg
        mermaid = build_cg(self.model_path)
        if save_path is None:
            # rfind: 找到最右边的index
            # 如果直接从左find，那么下面这个路径就找不到
            # ../AVNRT_Initial_straight.md
            save_path = self.model_path[: self.model_path.rfind(".")] + "_CG.md"
        with open(save_path, "w") as f:
            f.write(mermaid)
        return mermaid

    def verify(self, trace_path: str):
        """
        验证模型，并将验证结果保存到trace_path中
        """
        # 取出文件名../AVNRT_Initial_straight.xml
        idx = self.model_path.rfind('/')
        # tmp_model_path: ../tmp_verify_AVNRT_Initial_straight.xml
        tmp_model_path = f'{self.model_path[:idx + 1]}tmp_verify_{self.model_path[idx + 1:]}'
        self.save(tmp_model_path)
        # print(tmp_model_path)
        return Verifyta().simple_verify(tmp_model_path, trace_path)

    def get_templates(self):
        """
        根据名字获取相应template的Element
        template_name: string, template的名字
        """
        return self.element_tree.iter("template")

    def get_template(self, template_name: str):
        """
        根据名字获取相应template的Element
        template_name: string, template的名字
        """
        for template in self.element_tree.iter("template"):
            if template.find('name').text == template_name:
                return template
        return None

    def remove_template(self, template_name: str):
        """
        删除指定的template
        template_name: string
        if template is not found, return False.
        if template is successfully removed, return True.
        """
        template_elem = self.get_template(template_name)
        if template_elem is None:
            return False
        self.root_elem.remove(template_elem)
        return True
    
    # def same_name_template(self, template_name: str):


    def clear_queries(self):
        """
        删除queries_elem
        主要被set_queries调用
        """
        root = self.root_elem
        queries_elem = root.find('queries')
        if queries_elem is None:
            return False
        root.remove(queries_elem)
        return True

    def set_queries(self, queries: List[str]):
        """
        设置query
        queries: List[str], query组成的list
        返回修改queries后的self.get_queries()
        """
        # 首先删除所有的queries
        self.clear_queries()
        # 然后构造queries并插入到模型中
        queries_elem = UFactory.queries(queries)
        self.root_elem.append(queries_elem)
        return self.get_queries()

    def get_queries(self):
        """
        返回queries的字符串
        """
        query_formula_elems = self.element_tree.findall('./queries/query/formula')
        queries = ''.join([query_elem.text for query_elem in query_formula_elems])
        return queries

    def get_system(self):
        """
        返回system的字符串
        """
        system_elem = self.element_tree.find('system')
        return system_elem.text

    def set_system(self, system_str: str):
        """
        修改system的字符串为system_str
        """
        system_elem = self.element_tree.find('system')
        system_elem.text = system_str
        return system_str

    def add_system(self, system_str: str):
        """
        添加system。值得注意的是，这里实现方法是简单拼接system_str到末尾，并调整好末尾分号的位置，
        那么，如果想添加多个system，比如test1和test2，那么直接传入'test1, test2'即可
        返回self.get_system()
        """
        current_system = self.get_system()
        new_system = f'{current_system[:current_system.rfind(";")]},{system_str};'
        self.set_system(new_system)
        return self.get_system()

    def get_declaration(self):
        """
        返回declaration的字符串
        """
        declaration_elem = self.element_tree.find('declaration')
        return declaration_elem.text

    def set_declaration(self, declaration_str: str):
        """
        设置整个declarations
        """
        declaration_elem = self.element_tree.find('declaration')
        declaration_elem.text = declaration_str
        return declaration_str

    def add_declaration(self, declaration_str: str):
        """
        添加一条新的语句到最后一行或者第一行
        """
        current_declaration = self.get_declaration()
        new_declaration = f'{current_declaration[:-1]},{declaration_str};'
        self.set_declaration(new_declaration)
        return self.get_declaration()

    def get_max_location_id(self) -> int:
        """
        获取当前模型最大的location_id，方便制造新的模板
        """
        location_elems = self.element_tree.findall('./template/location')
        # <location id="id0" x="-187" y="-76">
        # <location id="id1" x="25" y="-76">
        # <location id="id2" x="-51" y="-119">
        ids = [int(location_elem.attrib['id'][2:]) for location_elem in location_elems]
        return max(ids)

    def get_broadcast_chan(self) -> List[str]:
        declarations = self.get_declaration()
        systems = self.get_system()
        start_index = 0
        broadcast_chan = []
        while True:
            start_index = declarations.find('broadcast chan', start_index, -1)
            if start_index == -1:
                break
            end_index = declarations.find(';', start_index, -1)
            tmp_actions = declarations[start_index+15:end_index].strip().split(',')
            tmp_actions = [x.strip() for x in tmp_actions]
            broadcast_chan += tmp_actions
            start_index = end_index
        start_index = 0
        # while True:
        #     start_index = systems.find('broadcast chan', start_index, -1)
        #     if start_index == -1:
        #         break
        #     end_index = systems.find(';', start_index, -1)
        #     tmp_actions = systems[start_index+15:end_index].strip().split(',')
        #     tmp_actions = [x.strip() for x in tmp_actions]
        #     broadcast_chan += tmp_actions
        #     start_index = end_index
        return list(set(broadcast_chan))
            

    def add_monitor(self, monitor_name: str, signals: TimedActions, observe_actions: List[str] = None, 
                    strict: bool = False, allpattern: bool = False):
        """
        在<system></system>前添加新的线性monitor，并且自动添加到system中
        如果name有冲突，会自动替换原template
        monitor_name: string
                 signals: List[Tuple[str, str, str]],
                 每个Tuple分别对应[signal, guard, inv]
                 example: ['act_path!', 'gclk>=10', 'gclk<=10']
                 signal, guard, inv, name 都可以是None
                 signal: 信号名称
                 guard: int
                 inv: int
                 注意：在连接的时候是按照信号在list出现的顺序连接的
        startID: int, 用来设置Monitor中最小的startID, 防止id冲突
        strict: bool, 判断是否构建strict monitor，正常来说是false
        """
        # 处理observe_actions is None的情况
        if observe_actions is None:
            observe_actions = self.get_broadcast_chan()

        start_id = self.get_max_location_id() + 1
        # 删除相同名字的monitor
        self.remove_template(monitor_name)
        monitor = UFactory.monitor(monitor_name, signals.convert_to_list_tuple(), observe_actions, start_id, strict, allpattern)
        self.root_elem.insert(-2, monitor)
        # 将新到monitor加入到system中
        self.add_system(monitor_name)
        return None

    def add_input(self, input_template_name: str, signals: TimedActions):
        """
        在<system></system>前添加新的线性monitor，并且自动添加到system中
        如果name有冲突，会自动替换原template
        monitor_name: string
               signals: List[Tuple[str, int, int]],
                 每个Tuple分别对应[signal, guard, inv]
                 signal, guard, inv, name 都可以是None
                 signal: 信号名称
                 guard: int
                 inv: int
                 注意：在连接的时候是按照信号在list出现的顺序连接的
        startID: int, 用来设置Monitor中最小的startID, 防止id冲突
        strict: bool, 判断是否构建strict monitor，正常来说是false
        """
        start_id = self.get_max_location_id() + 1
        # 删除相同名字的monitor
        self.remove_template(input_template_name)
        input_model = UFactory.input(input_template_name, signals.convert_to_list_tuple(), start_id)
        self.root_elem.insert(-2, input_model)
        # 将新到monitor加入到system中
        self.add_system(input_template_name)
        return None

    def find_a_pattern(self, inputs: TimedActions, observes: TimedActions, 
                       observe_actions: List[str] = None, focused_actions: List[str] = None, hold=False, options = None):
        """
        inputs: 输入信号模块的TimedActions
        observers: 观测信号模块的TimedActions
        input_actions: 输入信号列表
        observe_actions
        """
        # 设置路径
        # 新模型路径，不覆盖原模型
        new_model_path = os.path.splitext(self.model_path)[0] + '_pattern.xml'
        self.copy_as(new_model_path=new_model_path)
        new_umodel = UModel(new_model_path)
        # 将要保存的path路径
        # trace_path = f"{new_model_path.replace('.xml', '')}"

        patterns = []
        # 构建Monitor0

        new_umodel.add_input('input0', inputs)
        new_umodel.add_monitor('Monitor0', observes, observe_actions=observe_actions, strict=True)
        # 设置验证语句
        query = 'E<> Monitor0.pass'
        new_umodel.set_queries([query])
        # 保存构建好的模型
        new_umodel.save()
        # 获取第0个pattern
        Verifyta().simple_verify(new_model_path, options=options)
        trace_path = os.path.splitext(new_model_path)[0] + '-1.xtr'
        if not os.path.exists(trace_path):
            return []

        # 通过Trace 得到 Simtrace对象
        simtrace = Tracer.get_timed_trace(new_model_path, trace_path)
        # focused_actions = list(set(input_actions + hidden_actions + observe_actions))
        # print(focused_actions)
        pattern_seq = simtrace.filter_by_actions(focused_actions)

        if not hold:
            os.remove(new_model_path)
            os.remove(trace_path)
        return query, pattern_seq.actions

    def find_all_patterns(self, inputs: TimedActions, observes: TimedActions, 
                          observe_actions: List[str]=None, focused_actions: List[str]=None, hold: bool=False, max_patterns: int=None):
        """
        注意这里的input已经在模型中，并且原模型不包含任何Monitor

        observable_events: List[Tuple[str, str, str]]
        在给定input和observation的情况下寻找所有可能的counter example
        基本思路：
        Monitor0: observable events
        Monitor1: 基于Monitor0返回的trace
        """
        # 首先
        monitor_pass_str, new_patterns = self.find_a_pattern(inputs, observes, observe_actions, focused_actions, hold=True)
        new_model_path = os.path.splitext(self.model_path)[0] + '_pattern.xml'
        new_umodel = UModel(new_model_path)
        
        # 根据初始的pattern构建monitor并循环, 初始Moniter为0
        all_patterns = []
        monitor_id = 0
        iter = 1
        while len(new_patterns) != 0 and (iter <= max_patterns) if max_patterns is not None else True:
            monitor_id += 1
            all_patterns.append((monitor_pass_str, new_patterns))
            # 将pattern[List] -> TimedActions
            new_observes = TimedActions(new_patterns)
            new_umodel.add_monitor(f'Monitor{monitor_id}', new_observes, observe_actions=focused_actions, strict=True, allpattern=True)

            # 构造验证语句
            # 构造monitor.pass
            # !Monitor0.pass & !Monitor1.pass
            monitor_pass_str = ' && '.join([f'!Monitor{i}.pass' for i in range(1, monitor_id+1)])
            # E<> !Monitor0.pass & !Monitor1.pass
            monitor_pass_str = f'E<> Monitor0.pass && {monitor_pass_str}'

            # 设置验证语句
            new_umodel.set_queries([monitor_pass_str])
            # 保存构建好的模型
            new_umodel.save()
            
            Verifyta().simple_verify(new_umodel.model_path)
            
            trace_path = os.path.splitext(new_umodel.model_path)[0] + '-1.xtr'
            if not os.path.exists(trace_path):
                return []
            
            # 通过Trace 得到 Simtrace对象
            simtrace = Tracer.get_timed_trace(new_umodel.model_path, trace_path)
            new_patterns = simtrace.filter_by_actions(focused_actions).actions
            iter = iter + 1
        if not hold:
            os.remove(new_model_path)
            os.remove(trace_path)

        return all_patterns


    def find_a_pattern_with_query(self, query: str = None, focused_actions: List[str] = None, hold=False, options=None):
        """
        input_actions: 输入信号列表
        observe_actions
        """
        # 设置路径
        # 新模型路径，不覆盖原模型
        new_model_path = os.path.splitext(self.model_path)[0] + '_pattern.xml'
        self.copy_as(new_model_path=new_model_path)
        new_umodel = UModel(new_model_path)
        
        if query is not None:
            new_umodel.set_queries(queries=query)

        Verifyta().simple_verify(new_model_path,options=options)
        trace_path = os.path.splitext(new_model_path)[0] + '-1.xtr'
        if not os.path.exists(trace_path):
            return []
        
        # 通过Trace 得到 Simtrace对象
        simtrace = Tracer.get_timed_trace(self.model_path, trace_path)
        # focused_actions = list(set(input_actions + hidden_actions + observe_actions))
        # print(focused_actions)
        pattern_seq = simtrace.filter_by_actions(focused_actions)

        if not hold:
            os.remove(new_model_path)
            os.remove(trace_path)

        return new_umodel.get_queries(), pattern_seq.actions
    
    def find_all_patterns_with_query(self, query: str=None, focused_actions: List[str]=None, hold: bool=False, max_patterns: int=None):
        """
        注意这里的input已经在模型中，并且原模型不包含任何Monitor

        observable_events: List[Tuple[str, str, str]]
        在给定input和observation的情况下寻找所有可能的counter example
        基本思路：
        Monitor0: observable events
        Monitor1: 基于Monitor0返回的trace
        """
        # 首先
        query = query.strip()
        if not (query.startswith('A[]') or query.startswith('E<>')):
            raise NotImplementedError('Only support E<> and A[] query!')
        
        if query.startswith('A[]'):
            default_query = query[3:].strip()
            if default_query[0] == '!':
                default_query = "E<> " + default_query[1:].strip()
            elif default_query[0:3] == "not":
                default_query = "E<> " + default_query[3:].strip()
            else:
                default_query = "E<> ! " + default_query.strip()
        else:
            default_query = query
        default_query, new_patterns = self.find_a_pattern_with_query(default_query, focused_actions, hold=True)
        new_model_path = os.path.splitext(self.model_path)[0] + '_pattern.xml'
        new_umodel = UModel(new_model_path)
        monitor_pass_str = default_query
        # 根据初始的pattern构建monitor并循环, 初始Moniter为0
        all_patterns = []
        monitor_id = 0
        iter = 1
        while len(new_patterns) != 0:
            all_patterns.append((monitor_pass_str, new_patterns))
            if max_patterns is not None and iter >= max_patterns:
                break
            monitor_id += 1
            # 将pattern[List] -> TimedActions
            new_observes = TimedActions(new_patterns)
            new_umodel.add_monitor(f'Monitor{monitor_id}', new_observes, observe_actions=focused_actions, strict=True, allpattern=True)

            # 构造验证语句
            # 构造monitor.pass
                # E<> Monitor0.pass & !Monitor1.pass
            monitor_pass_str = ' && '.join([f'!Monitor{i}.pass' for i in range(1, monitor_id+1)])
            # E<> !Monitor0.pass & !Monitor1.pass
            monitor_pass_str = f'{default_query} && {monitor_pass_str}'
                
            
            # 设置验证语句
            new_umodel.set_queries([monitor_pass_str])
            # 保存构建好的模型
            new_umodel.save()
            
            Verifyta().simple_verify(new_umodel.model_path)
            
            trace_path = os.path.splitext(new_umodel.model_path)[0] + '-1.xtr'
            if not os.path.exists(trace_path):
                return []
            
            # 通过Trace 得到 Simtrace对象
            simtrace = Tracer.get_timed_trace(new_umodel.model_path, trace_path)
            new_patterns = simtrace.filter_by_actions(focused_actions).actions
            iter = iter + 1
        if not hold:
            os.remove(new_model_path)
            os.remove(trace_path)
        return all_patterns


from typing import List
from collections import namedtuple


# timedtuple
class TimedActions:
    def __init__(self, actions: List[str], lb: List[int] = None, ub: List[str] = None):
        self.actions = actions
        self.lb = lb
        self.ub = ub
        if self.lb is None:
            self.lb = [-1 for i in range(len(self.actions))]
        if self.ub is None:
            self.ub = [-1 for i in range(len(self.actions))]
    
    @property
    def is_patterns(self):
        return all(self.lb) == -1 and all(self.ub) == -1

    def convert_to_list_tuple(self, clk_name='monitor_clk'):
        self.clk_name = clk_name
        res = []
        for i in range(len(self.actions)):
            res.append((self.actions[i], f'{clk_name}>={self.lb[i]}', f'{clk_name}<={self.ub[i]}'))
        return res
    
    def convert_to_patterns(self):
        return self.actions



from .verifyta import Verifyta
from .umodel import UModel
from .tracer import ClockZone, Transition, SimTrace, Tracer
from .config import *
from .pyuppaal import *
from .namedtuple import *



# __all__ = [ROOT_DIR, TRACER_CUSTOM_WINDOWS,TRACER_CUSTOM_LINUX]



# 这一行的import能够指定class的method返回自身类
# 参考链接：https://www.nuomiphp.com/eplan/11188.html
from __future__ import annotations
from typing import List, Tuple, Dict
import xml.etree.cElementTree as ET
from .verifyta import Verifyta
from .config import *
import os
import platform


platform_system = platform.system()
tracer_custom = TRACER_CUSTOM_WINDOWS if platform.system() == 'Windows' else TRACER_CUSTOM_LINUX


# ClockZone = namedtuple('ClockZone', ['clock1', 'clock2', 'is_equal', 'bound_value'])
# Transition = namedtuple('Transition', ['sync', 'start_process', 'end_process'])

class OneClockZone:
    def __init__(self, clock1: str, clock2: str, is_equal: bool, bound_value: int):
        """
        clock1 - clock2 < or ≤ bound_value
        """
        self.clock1: str = clock1
        self.clock2: str = clock2
        self.is_equal: bool = is_equal
        self.bound_value: int = bound_value

    def __str__(self):
        sign = '≤' if self.is_equal else '<'
        res = f'{self.clock1} - {self.clock2} {sign} {self.bound_value}'
        return res


class ClockZone:
    def __init__(self, clockzones: List[OneClockZone]):
        """
        ClockZone List
        """
        self.clockzones = clockzones

    def __str__(self):
        res = f'[]'
        if len(self.clockzones) > 0:
            res = '['
            for i in range(len(self.clockzones)):
                res += f'{self.clockzones[i].__str__()}; '
            res += ']'
        return res


class Edges:
    def __init__(self, start_location, end_location, Guard: str, Sync: str, Update: str):
        """Edges
        process.start_location -> process.end_location: {Guard, Sync, Update} 
        """
        self.start_location = start_location
        self.end_location = end_location
        self.Guard = Guard
        self.Sync = Sync
        self.Update = Update
        self.process = self.start_location.split('.')[0]
        # self.is_sync = Sync != '0'
        # self.is_grard = Guard != '1'

    def __str__(self):
        res = f'{self.process}.{self.start_location} -> {self.process}.{self.end_location}:'
        res = res + f'{{{self.Guard};{self.Sync};{self.Update}}}'
        return res

    @property
    def is_sync(self):
        return self.Sync[-1] == '!' or self.Sync[-1] == '?'

    @property
    def sync_symbol(self):
        if self.is_sync:
            return self.Sync[:-1]
        else:
            return None

    @property
    def is_guard(self):
        return self.Guard != '1'

    @property
    def sync_type(self):
        if self.is_sync:
            return 'send' if self.Sync[-1] == '!' else 'receive'
        else:
            return None


class Transition:
    def __init__(self, sync: str, start_process: str, end_process: List[str], edges: List[Edges] = None):
        self.sync: str = sync
        self.start_process: str = start_process
        self.end_process: List[str] = end_process
        # 这个edges暂时没用，但是我们也先存下来了。
        self.edges: List[Edges] = edges

    def __str__(self):
        if self.sync is None:
            res = f'{self.sync}: {self.edges[0].start_location} -> {self.edges[0].end_location}'
        else:
            res = f'{self.sync}: {self.start_process} -> {self.end_process}'
        return res

    @property
    def action(self):
        return self.sync


class GlobalVar:
    def __init__(self, variables_name: List[str], variables_value: List[float]):
        """
        GlobalVar: variables_name[i] = variables_value[i]
        """
        self.variables_name = variables_name
        self.variables_value = variables_value

    def __str__(self):
        res = f'[]'
        if len(self.variables_name) > 0:
            res = '['
            for i in range(len(self.variables_name)):
                res += f'{self.variables_name[i]}={self.variables_value[i]}; '
            res += ']'
        return res


class SimTrace:
    def __init__(self, states: List[List[str]] = [],
                 clock_constraints: List[ClockZone] = [],
                 transitions: List[Transition] = [],
                 global_variables: List[GlobalVar] = []):
        """
        第i个state和第i个transition有相同的clock constraints
        第i个transition前面跟第i个state，详情看__str__()
        """
        self.__states: List[List[str]] = states
        self.__global_variables = global_variables
        self.__clock_constraints: List[ClockZone] = clock_constraints
        self.__transitions: List[Transition] = transitions

        # check length correct
        # is_equal_conlen = len(self.__states) == len(self.__clock_constraints)
        # is_equal_tralen = len(self.__states) == len(
        #     self.__transitions) + 1 or (len(self.__states) == len(self.__transitions) == 0)
        # if not (is_equal_conlen and is_equal_tralen):
        #     raise ValueError(f'length should satisfy: len(states) == len(clock_constraints) == len(transitions) + 1\n'
        #                      f'current length: len(states) = {len(self.__states)}, '
        #                      f'len(clock_constraints) = {len(self.__clock_constraints)}, '
        #                      f'len(transitions) + 1 = {len(self.__transitions)+1}.')

    def __str__(self):
        res = ''
        for i in range(len(self.__states)):
            res += f'State [{i}]: {self.__states[i].__str__()}\n'
            res += f'global_variables [{i}]: {self.__global_variables[i].__str__()}\n'
            res += f'Clock_constraints [{i}]: {self.__clock_constraints[i].__str__()}\n'
            if i < len(self.__states) - 1:
                res += f'transitions [{i}]: {self.__transitions[i].__str__()}\n'
                res += f'-----------------------------------\n'
        return res

    def set_empty(self):
        self.__states: List[List[str]] = []
        self.__global_variables = []
        self.__clock_constraints: List[ClockZone] = []
        self.__transitions: List[Transition] = []

    @property
    def states(self):
        return self.__states

    @property
    def clock_constraints(self):
        return self.__clock_constraints

    @property
    def transitions(self):
        return self.__transitions

    # 这个就是get_untimed_pattern
    @property
    def actions(self) -> List[str]:
        return [x.action for x in self.transitions if x.action is not None]

    def filter_by_index(self, index_array: List[int]) -> SimTrace:
        """通过下标索引筛选出对应的SimTrace

        Args:
            index_array (List[int]): 索引

        Returns:
            SimTrace: 返回一个SimTrace对象
        """
        new_states = [self.__states[i] for i in range(len(self.__states)) if i in index_array]
        new_clock_cons = [self.__clock_constraints[i] for i in range(len(self.__clock_constraints)) if i in index_array]
        new_transitions = [self.__transitions[i] for i in range(len(self.__transitions)) if i in index_array]
        new_global_var = [self.__global_variables[i] for i in range(len(self.__global_variables)) if i in index_array]
        new_simtrace = SimTrace(new_states, new_clock_cons, new_transitions, new_global_var)
        return new_simtrace

    def filter_by_actions(self, focused_actions: List[str]=None) -> SimTrace:
        """
        filter edges by actions
        return 保留edges以及clock constraints
        """
        if focused_actions is None:
            return self
        
        index_array = [i for i in range(len(self.transitions)) if self.transitions[i].action in focused_actions]
        return self.filter_by_index(index_array)


    def filter_by_clocks(self, concerned_clocks: List[str], is_both: bool) -> SimTrace:
        """通过clockZone筛选trace

        Args:
            concerned_clocks (List[str]): _description_
            is_both (bool): _description_

        Returns:
            SimTrace: _description_
        """
        if is_both:
            pass
        else:
            pass
        pass
        raise NotImplementedError('NotImplementedError')

    def get_untime_pattern(self):
        return self.actions
    
    # def get_timed_pattern(self):
        
        


class Tracer:
    """
    用来分析由命令行验证得到的xtr格式的trace
    """

    @staticmethod
    def get_timed_trace(model_path: str, trace_path: str, hold: bool=False) -> SimTrace:
        """
        获取trace_path的trace以及每个转移状态对应的gclk的区间
        注意由于UPPAAL验证返回的限制，这个trace并不是具体的时间，而是gclk的限制
        能够帮助我们构建更快速验证的Monitor，但是不一定能帮我们划分参数
        返回 一个 SimulationTrace class.

        hold: 保留中间生成文件(.if & .txt)
        """
        verifyta = Verifyta()
        # use Verifta to compile model_path to if format file
        if_file = verifyta.compile_to_if(model_path=model_path)
        file_path, file_ext = os.path.splitext(if_file)
        # use trcer_custom to generate txt file
        trace_txt = file_path+'.txt'
        cmd_command = f'{tracer_custom} {if_file} {trace_path} {trace_txt}'
        cmd_res = os.popen(cmd_command).read()
        # check file exist
        if not os.path.exists(trace_txt):
            error_info = f'trace txt file {trace_txt} has not generated.'
            raise FileNotFoundError(error_info)
        # load trace txt file
        f = open(trace_txt, 'r')
        trace_text = f.readlines()
        f.close()

        # construct SimulationTrace instance
        clock_constraints, states, global_variables, transitions = [], [], [], []
        for tr_ind in range(len(trace_text)):
            state_text, globalvar_name, globalvar_val, clockzones_text, transitions_text = [], [], [], [], []
            if trace_text[tr_ind].startswith('State'):
                tmp_trace_i = trace_text[tr_ind][7:].strip().split(' ')
                # print(trace_text[tr_ind][7:].strip())
                for tr_sub in range(len(tmp_trace_i)):
                    if ("=" in tmp_trace_i[tr_sub]) and ('<' not in tmp_trace_i[tr_sub]):
                        # globalvariables
                        var_name, var_val = tmp_trace_i[tr_sub].split('=')
                        globalvar_name.append(var_name.strip())
                        globalvar_val.append(var_val.strip())
                    elif ('<' in tmp_trace_i[tr_sub]) and ('-' in tmp_trace_i[tr_sub]):
                        # clockzones
                        # print(tmp_trace_i[tr_sub])
                        clocks, clk_bound = [x.strip() for x in tmp_trace_i[tr_sub].split('<')]
                        clock1, clock2 = [x.strip() for x in clocks.split('-')]
                        if clk_bound[0] == '=':
                            clk_bound = clk_bound[1:]
                            is_equal = True
                        else:
                            is_equal = False
                        clkzone = OneClockZone(clock1=clock1.strip(), clock2=clock2.strip(' '),
                                            is_equal=is_equal, bound_value=clk_bound.strip(' '))
                        clockzones_text.append(clkzone)
                    else:
                        # state
                        state_text.append(tmp_trace_i[tr_sub].strip())
                clock_constraints.append(ClockZone(clockzones_text))
                states.append(state_text)
                global_variables.append(GlobalVar(
                    globalvar_name, globalvar_val))
                # print(clockzones_text[0].__str__())
                # print(state_text)
                # print(globalvar_name, globalvar_val)
            elif trace_text[tr_ind].startswith('Transition'):
                tmp_trace_i = trace_text[tr_ind][12:].strip().split('}')[:-1]
                edges_list = []
                end_process = []
                sync_symbol = None
                start_process = None
                for tr_sub in range(len(tmp_trace_i)):
                    trans_comp, edge_trans = tmp_trace_i[tr_sub].split('{')
                    start_location, end_location = [x.strip() for x in trans_comp.split('->')]
                    guard, sync, update = [x.strip() for x in edge_trans.split(';')[:-1]]
                    # strip
                    tmp_edge = Edges(
                        start_location, end_location, guard, sync, update)
                    if tmp_edge.sync_type == 'send':
                        start_process = tmp_edge.process
                        sync_symbol = tmp_edge.sync_symbol
                    elif tmp_edge.sync_type == 'receive':
                        end_process.append(tmp_edge.process)
                    else:
                        start_process = tmp_edge.process
                        end_process.append(tmp_edge.process)
                    edges_list.append(tmp_edge)
                transitions.append(Transition(sync=sync_symbol, start_process=start_process,
                                              end_process=end_process, edges=edges_list))
            else:
                pass
        # remove if and txt file
        if not hold:
            os.remove(if_file)
            os.remove(trace_txt)
        return SimTrace(states, clock_constraints, transitions, global_variables)

    @staticmethod
    def get_untime_pattern(trace_path: str, edge_signal_dict: Dict[str, str]):
        """
        获取trace_path的untime的pattern，即所有的!actions，并根据edge_signal_dict替换成可读信号
        trace_path: str, xml trace的路径
        edge_signal_dict示例：{'NSA.Edge2.actPath!': 'actPathSA!'}
        意思是将'NSA.Edge2.actPath!'替换为'actPathSA'
        注意，为了方便构造edge_signal_dict，可以调用并打印Tracer.get_untime_trace()
        """
        

        untime_trace = Tracer.get_untime_trace(trace_path)
        # print('========== get_untime_pattern', trace_path, edge_signal_dict, untime_trace)
        return Tracer.convert_trace_to_pattern(untime_trace, edge_signal_dict)




    @staticmethod
    def get_timed_pattern(trace_path: str) -> List[Tuple[str, str, str]]:
        """
        利用convert_trace_to_pattern函数将trace转化为易读的pattern
        能够帮助我们构建更快速验证的Monitor，但是不一定能帮我们划分参数
        """
        pass

    @staticmethod
    def get_transition_clock_bounds(trace_path: str) -> List[Dict[str, Dict[str, Tuple[str, int]]]]:
        """
        获取trace_path的clock_bounds
        trace_path: trace的xml路径
        return clock_bounds的列表
                clock_bounds里的每个Dict代表着（key减value）的取值范围
                比如 {'clock A':{'clock B':["<=",5]}}的意思是
                'clock A-clock B <= 5'

                clock_bound['clock A']['clock B'] = ('<=', 5)
        """
        et = ET.ElementTree(file=trace_path)

        nodes = list()
        for node in et.iter('node'):
            # node.attrib 示例
            # {'id': 'State42', 'location_vector': 'LocVec42', 'dbm_instance': 'DBM42 ', 'variable_vector': 'varvec.1'}
            # attrib_id = node.attrib['id']
            nodes.append(node.attrib['dbm_instance'][:-1])
        # print(nodes)

        # 将dbm按照"DBM1","DBM2"的顺序进行排序
        dbms = []
        i = 1
        while nodes:
            dbm = 'DBM' + str(i)
            if nodes.count(dbm) == 0:
                i += 1
                continue
            else:
                index = nodes.index(dbm)
                dbms.append(nodes[index])
                nodes.pop(index)
        # print(dbms)

        # 寻找所有的dbm_instance
        dbm_instances = {}
        for dbm_instance in et.iter('dbm_instance'):
            dbm_instances[dbm_instance.attrib['id']] = dbm_instance

        # 将所有的dbm_instance的信息按照dbm的顺序进行储存
        transition_clock_bounds = []
        for dbm in dbms:
            dbm_instance = dbm_instances[dbm]
            transition_clock_bound = dict()
            for child in dbm_instance:
                attrib = child.attrib
                # {'clock1': 'NHisV.t', 'clock2': 'NHisV.t', 'bound': '0', 'comp': '<='}
                key_a = attrib['clock1']
                key_b = attrib['clock2']
                val = (attrib['comp'], int(attrib['bound']))
                if key_a in transition_clock_bound:
                    transition_clock_bound[key_a].update({key_b: val})
                else:
                    transition_clock_bound.update({key_a: {key_b: val}})

            transition_clock_bounds.append(transition_clock_bound)

        return transition_clock_bounds

    @staticmethod
    def get_edge_clock_bounds(trace_path: str) -> [Dict[str, List[Tuple[Tuple[str, int], Tuple[str, int]]]]]:
        """
        获取trace_path的edge对应的clock_bounds
        trace_path: trace的xml路径
        return 每条边对应的clock_bounds的列表
                edge对应的clock_bounds里的每个Dict代表着经过key的时间的取值范围
                比如{'PSAHisA.Edge2': [(('>=', 0), ('<=', 28)), (('>=', 0), ('<=', 28))]}的意思是
                '经过PSAHisA.Edge2两次，第一次时间t>=0且t<=28, 第二次t>=0且t<=28'
        """

        et = ET.ElementTree(file=trace_path)

        # 为get_transition_clock_bounds函数
        nodes = list()
        for node in et.iter('node'):
            # node.attrib 示例
            # {'id': 'State42', 'location_vector': 'LocVec42', 'dbm_instance': 'DBM42 ', 'variable_vector': 'varvec.1'}
            # attrib_id = node.attrib['id']
            nodes.append(node.attrib['dbm_instance'][:-1])
        # print(nodes)

        # 将dbm按照"DBM1","DBM2"的顺序进行排序
        dbms = []
        i = 1
        while nodes:
            dbm = 'DBM' + str(i)
            if nodes.count(dbm) == 0:
                i += 1
                continue
            else:
                index = nodes.index(dbm)
                dbms.append(nodes[index])
                nodes.pop(index)
        # print(dbms)

        # 寻找所有的dbm_instance
        dbm_instances = {}
        for dbm_instance in et.iter('dbm_instance'):
            dbm_instances[dbm_instance.attrib['id']] = dbm_instance

        # 将所有的dbm_instance的信息按照dbm的顺序进行储存
        transition_clock_bounds = []
        for dbm in dbms:
            dbm_instance = dbm_instances[dbm]
            transition_clock_bound = dict()
            for child in dbm_instance:
                attrib = child.attrib
                # {'clock1': 'NHisV.t', 'clock2': 'NHisV.t', 'bound': '0', 'comp': '<='}
                key_a = attrib['clock1']
                key_b = attrib['clock2']
                val = (attrib['comp'], int(attrib['bound']))
                if key_a in transition_clock_bound:
                    transition_clock_bound[key_a].update({key_b: val})
                else:
                    transition_clock_bound.update({key_a: {key_b: val}})

            transition_clock_bounds.append(transition_clock_bound)

        #     寻找transition对应的发出信号的边
        edge_clock_bounds = dict()
        count = 0
        # print(trace_path)
        for transition in et.iter('transition'):
            # transition.attrib 示例
            # {'from': 'State1', 'to': 'State2', 'edges': 'Input.Edge1 ConvertIn.Edge1 '}
            # print(transition.attrib)
            edge = transition.attrib['edges'].split(' ')[0]
            # print(edge)
            node = edge.split('.')[0]
            nodet = node + '.t'
            if nodet in transition_clock_bounds[count]:
                # print(edge)
                # print(transition_clock_bounds[count][nodet]['sys.t(0)'][1])
                # print(transition_clock_bounds[count]['sys.t(0)'][nodet][1])
                small = transition_clock_bounds[count]['sys.t(0)'][nodet]
                large = transition_clock_bounds[count][nodet]['sys.t(0)']
                small = (small[0].replace('<', '>'), small[1])
                value = (small, large)
                # print(value)
                if edge not in edge_clock_bounds:
                    edge_clock_bounds[edge] = [value]
                else:
                    edge_clock_bounds[edge].append(value)
            count += 1

        return edge_clock_bounds

    @staticmethod
    def get_parameter_edge(trace_path: str) -> Dict[str, List[str]]:
        """
        {'tERPMinSA': ['NSA.Edge3'], 'tCondMinSAHisA': ['PSAHisA.Edge2'], 'tERPMinHisA': ['NHisA.Edge3'],
        'tCondMinHisASP': ['PHisASP.Edge4'], 'tCondMinHisAFP': ['PHisAFP.Edge2'], 'tERPMinSP': ['NSP.Edge3'],
        'tERPMinFP': ['NFP.Edge4'], 'tCondMinSPHisH': ['PSPHisH.Edge2', 'PSPHisH.Edge5'], 'tCondMinFPHisH': [
        'PFPHisH.Edge2'], 'tERPMinHisH': ['NHisH.Edge3'], 'tCondMinHisHHisV': ['PHisHHisV.Edge2'], 'tERPMinHisV': [
        'NHisV.Edge3']}
        """
        parameter_edge = dict()
        et = ET.ElementTree(file=trace_path)
        for edge in et.iter('edge'):
            # edge.attrib 示例
            # {'id': 'NSA.Edge1', 'from': 'NSA.Rest', 'to': 'NSA._id2'}

            # from xml.etree.ElementTree import Element, tostring, fromstring
            # print(tostring(edge, encoding="unicode"))

            for guard in edge.iter('guard'):
                # t >= tERPMin
                text = guard.text
                if 't >= t' in guard.text:
                    edge = edge.attrib['id']
                    # print(edge.attrib['id'])
                    # NSA.Edge3

                    parameter = 't' + guard.text.split('t >= t')[1]
                    # print(guard.text.split('t >= t')[1])
                    # ERPMin

                    # ERPMin += SA
                    # parameter = ERPMinSA
                    parameter += edge.split('.')[0][1:]

                    if parameter in parameter_edge:
                        parameter_edge[parameter].append(edge)
                    else:
                        parameter_edge[parameter] = [edge]

        return parameter_edge

    @staticmethod
    def get_untime_trace(trace_path: str):
        """
        这里我们不调用get_timed_trace，因为可以减少对DBM的搜索，从而提高运行效率
        获取trace_path的untime的trace
        trace_path: trace的xml路径
        return transitions的列表
                transitions里的每个list的第0个元素是发送edge，后面的为接收edge
                比如['NSA.Edge2', 'PSAHisA.Edge1', 'Monitor.Edge1']的意思是
                'NSA.Edge2'发送信号给了'PSAHisA.Edge1'和'Monitor.Edge1'
        """
        et = ET.ElementTree(file=trace_path)
        # print('======== get_untime_trace', ET.dump(et))
        # 找到trace中所有的edges，方便为transitions的edges添加后缀
        # 比如'NSA.Edge1'添加'.actNode?'后缀变为'NSA.Edge1.actNode?'
        # edges长这样：
        '''
        <edge id="Monitor.Edge1" from="Monitor._id44" to="Monitor._id33">
                <guard>gclk &gt;= 0</guard>
                <sync>actPathSA?</sync>
                <update>1</update>
        </edge>
        '''
        # 构建edge: sync的字典
        # 例如：{'NSA.Edge1': 'actNode?', 'NSA.Edge2': 'actPath!'}
        edges = {}
        for edge in et.iter('edge'):
            # edge.attrib 示例  {'id': 'NSA.Edge1', 'from': 'NSA.Rest', 'to': 'NSA._id2'}
            attrib_id = edge.attrib['id']
            edges[attrib_id] = f"{attrib_id}.{edge.find('sync').text}"

        # 找到trace中所有的transitions
        # transitions长这样：
        '''
        <transition from="State1" to="State2" edges="Input.Edge1 ConvertIn.Edge1 "/>
        <transition from="State2" to="State3" edges="ConvertIn.Edge2 NSA.Edge1 "/>
        <transition from="State3" to="State4" edges="NSA.Edge2 PSAHisA.Edge1 Monitor.Edge1 "/>
        '''
        # transitions例子：['Input.Edge1', 'ConvertIn.Edge1', 'ConvertIn.Edge2']
        transitions = []
        for transition in et.iter('transition'):
            # transition.attrib: {'from': 'State1', 'to': 'State2', 'edges': 'Input.Edge1 ConvertIn.Edge1 '}
            tmp = transition.attrib['edges'].split(' ')
            tmp.pop()
            # 为transitions加后缀，比如
            # [['Input.Edge1', 'ConvertIn.Edge1'], ['ConvertIn.Edge2', 'NSA.Edge1']]变为
            # [['Input.Edge1.sigIn!', 'ConvertIn.Edge1.sigIn?'], ['ConvertIn.Edge2.actNodeSA!', 'NSA.Edge1.actNode?']]
            for i in range(len(tmp)):
                tmp[i] = edges[tmp[i]]
            transitions.append(tmp)

        # print(edges)
        return transitions, edges

    @staticmethod
    def convert_trace_to_pattern(untime_trace: List[List[str]], edge_signal_dict: Dict[str, str]):
        # 改为
        # convert_trace_to_pattern(trace: List[List[str, str, str]], edge_signal_dict: Dict[str, str])
        # -> List[List[str, str, str]]
        # 因为我们的pattern、trace等都定义为timed words
        """
        根据edge_signal_dict将Tracer.get_untime_trace()得到的untime_trace转化成可读信号
        edge_signal_dict示例：{'NSA.Edge2.actPath!': 'actPathSA!'}
        意思是将'NSA.Edge2.actPath!'替换为'actPathSA'
        """
        if edge_signal_dict is None:
            print('self.edge_signal_dict is None, untime_trace is: ')
            print(untime_trace)
        # pattern = [i[0] for i in untime_trace if 'tau' not in i[0]]
        # print(untime_trace)

        # pattern = [edge_signal_dict[i[0]] for i in untime_trace if i[0] in edge_signal_dict]
        pattern = [edge_signal_dict[i[0]]
                   for i in untime_trace[0] if i[0] in edge_signal_dict]
        return pattern

    @staticmethod
    def validate_and_get_untime_pattern(model_path: str, trace_path: str, edge_signal_dict: Dict[str, str]):
        """
        验证并返回pattern
        model_path: str, 要验证的模型路径
        trace_path: str, 准备保存的trace路径
        """
        verify_res = Verifyta().simple_verify(model_path, trace_path)
        # print('validate_and_get_untime_pattern', verify_res)
        if 'Formula is satisfied' in verify_res:
            # 要创建一个新的trace path，因为如果trace_path是test.xml
            # 那么验证得到的模型应该是test1.xml
            trace_path = trace_path.replace('.xml', '1.xml')
            new_trace_path = trace_path
            pattern = Tracer.get_untime_pattern(
                new_trace_path, edge_signal_dict)
            # print('======== validate_and_get_untime_pattern', pattern, model_path, trace_path)
            return pattern
        else:
            return []

    @staticmethod
    def get_old_pattern_obs_events(trace_path, edge_signal_dict):
        trace_path = trace_path.replace('.xml', '1.xml')
        trace = Tracer.get_untime_trace(trace_path)[0]
        transition_clock_bounds = Tracer.get_transition_clock_bounds(
            trace_path)
        old_pattern_obs_events = []
        for i in range(len(trace)):
            transition = trace[i]
            if transition[0] in edge_signal_dict:
                signal = edge_signal_dict[transition[0]]
                signal = signal.replace('!', '?')
                transition_clock_bound = transition_clock_bounds[i]
                small = -transition_clock_bound['sys.t(0)']['sys.gclk'][1]
                large = transition_clock_bound['sys.gclk']['sys.t(0)'][1]
                old_pattern_obs_events.append(
                    (signal, 'gclk>=' + str(small), 'gclk<=' + str(large)))
        return old_pattern_obs_events

    @staticmethod
    def extract_simulation_trace(xtr_file: str) -> SimulationTrace:
        pass
